#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stddef.h>
#include <signal.h>
#include <math.h>

#include <X11/Xdefs.h>
#include <X11/Xlib.h>
#include "elantech.h"


enum ParaType {
    PT_INT,
    PT_BOOL,
    PT_DOUBLE
};

struct Parameter {
    char *name;                             /* Name of parameter */
    int offset;                             /* Offset in shared memory area */
    enum ParaType type;                     /* Type of parameter */
    double min_val;                         /* Minimum allowed value */
    double max_val;                         /* Maximum allowed value */
};

#define DEFINE_PAR(name, memb, type, min_val, max_val) \
{ name, offsetof(ElantechSHM, memb), (type), (min_val), (max_val) }

static struct Parameter params[] = {
	DEFINE_PAR("DetectCovergestureIntr",          	detect_cover_intr,            	PT_BOOL,    0, 1),
	{ 0, 0, 0, 0, 0 }
};

static int keyboard_activity(Display *display)
{
    unsigned char key_state[KEYMAP_SIZE];
    int i;
    int ret = 1;

    XQueryKeymap(display, (char*)key_state);
    ret=0;

   
    if(key_state[6]==4  || /*left shift*/ 
       key_state[4]==32 || /*left ctrl*/
       key_state[8]==1  || /*left alt*/
       key_state[14]==8 || /*left windowns key*/
       key_state[7]==64 || /*right shift*/
       key_state[13]==32 || /*right ctril*/
       key_state[14]==2 || /*right alt*/
       key_state[8]==4 || /*caps lock*/
       key_state[12]==4 ||/*up*/
       key_state[13]==1 ||/*down*/
       key_state[12]==16 ||/*left*/
       key_state[12]==64 ||/*right*/
       key_state[13]==2  || /*PgUp*/
       key_state[13]==8  ||   /*PgDn*/
       key_state[1]==2      /*ESC*/
      )
     {
	ret=0;
     }
    else
    {
	for (i = 0; i < KEYMAP_SIZE; i++)
	{
		if(key_state[i] != 0)
			break;
	}
	
	if(i >= KEYMAP_SIZE)
		ret=1;
	else
	{
		//printf("key_state[%d]=%d\n",i,key_state[i]);
		ret=2; 
	}
    }

    //printf("ret=%d\n",ret);
    return ret;
}


int main()
{

	ElantechSHM *elashm;
	int shmid;
	int ret;
	int inter_second_time=0;
	int special_key;
	int sleep_time;
	double value;
	Display *display;

	display = XOpenDisplay(NULL);
	if(!display)
	{
		fprintf(stderr,"Can't open display.\n");
		exit(2);
	}

	/* Connect to the shared memory area */
    	if ((shmid = shmget(SHM_ELANTECH, sizeof(ElantechSHM), 0)) == -1) {
        	if ((shmid = shmget(SHM_ELANTECH, 0, 0)) == -1) {
            		fprintf(stderr, "Can't access shared memory area. SHMConfig disabled?\n");
            		exit(1);
        	} else {
            		fprintf(stderr, "Incorrect size of shared memory area. Incompatible driver version?\n");
            		exit(1);
        	}
    	}
    	if ((elashm = (ElantechSHM*) shmat(shmid, NULL, 0)) == NULL) {
        	perror("shmat");
        	exit(1);
    	}
	
	value=0;	
  	while(1) {
     		
		//printf("ret=%d\n",ret);
     		//(ret == 0) ? printf("\n\rInterrupt occurs") : printf("\n\rNo interrupt occurs");
		struct Parameter* par = &params[0];
		if(elashm->disable_when_key_checkBox ){
			special_key = keyboard_activity(display);	
			if(special_key  != 1){	
				//ret = system("/usr/local/bin/.KbdInterDetect.sh 0.1");
				if(special_key==2)
				{
					if(value==0){
						value=1;
		 				*(Bool*)((char*)elashm + par->offset) = (Bool)rint(value);
					}
					sleep_time = elashm->cover_delay_comboBox*100000;;
					usleep(sleep_time);
					//usleep(sleep_time);
					inter_second_time=0;
					
				}else{
					if(value==1){
						value=0;
		 				*(Bool*)((char*)elashm + par->offset) = (Bool)rint(value);
					}
					inter_second_time=1;
					
				}
				
			}
			else
			{
				if(inter_second_time < 10000)
				{
					usleep(10);
					inter_second_time++;
					if(inter_second_time == 1){
						usleep(sleep_time);
						if(value==1){
							value=0;
							*(Bool*)((char*)elashm + par->offset) = (Bool)rint(value);
						}
					}else if(inter_second_time > 1){
						if(value==1){
							value=0;
							*(Bool*)((char*)elashm + par->offset) = (Bool)rint(value);
						}
					}
				}
				else
				{
					sleep(1);
				}
			}
		}else{
			
			if(value==1)
			{
				value=0;
				*(Bool*)((char*)elashm + par->offset) = (Bool)rint(value);
			}
			sleep(1);
		}
		
  	}

  	return 0;
}