/*   Elantech Passthrough Support
 *   Copyright (c) 2008 Elantech Inc Tom Lin<tom_lin@elantech.com.tw>
 *   
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Trademarks are the property of their respective owners.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <stddef.h>
#include <math.h>

#include <X11/Xdefs.h>
#include "elantech.h"
#define ELAN_VERSION "ETD7.1.4 Free"
enum ParaType {
    PT_INT,
    PT_BOOL,
    PT_DOUBLE
};

struct Parameter {
    char *name;                             /* Name of parameter */
    int offset;                             /* Offset in shared memory area */
    enum ParaType type;                     /* Type of parameter */
    double min_val;                         /* Minimum allowed value */
    double max_val;                         /* Maximum allowed value */
};

#define DEFINE_PAR(name, memb, type, min_val, max_val) \
{ name, offsetof(ElantechSHM, memb), (type), (min_val), (max_val) }


static struct Parameter params[] = {
    DEFINE_PAR("MaxTapTime",           tap_time,                PT_INT,    0, 1000),
    DEFINE_PAR("MaxTapMove",           tap_move,                PT_INT,    0, 2000),
    DEFINE_PAR("MaxDoubleTapTime",     tap_time_2,              PT_INT,    0, 1000),
    DEFINE_PAR("SingleTapTimeout",     single_tap_timeout,      PT_INT,    0, 1000),
    DEFINE_PAR("ClickTime",            click_time,              PT_INT,    0, 1000),
    DEFINE_PAR("EmulateMidButtonTime", emulate_mid_button_time, PT_INT,    0, 1000),
    DEFINE_PAR("VertScrollDelta",      scroll_dist_vert,        PT_INT,    0, 1000),
    DEFINE_PAR("HorizScrollDelta",     scroll_dist_horiz,       PT_INT,    0, 1000),
    DEFINE_PAR("MinSpeed",             min_speed,               PT_DOUBLE, 0.05, 0.75),
    DEFINE_PAR("MaxSpeed",             max_speed,               PT_DOUBLE, 0.025, 0.75),
    DEFINE_PAR("AccelFactor",          accl,                    PT_DOUBLE, 0.00001, 0.2),
    DEFINE_PAR("UpDownScrolling",      updown_button_scrolling, PT_BOOL,   0, 1),
    DEFINE_PAR("LeftRightScrolling",   leftright_button_scrolling, PT_BOOL,   0, 1),
    DEFINE_PAR("UpDownScrollRepeat",   updown_button_repeat,    PT_BOOL,   0, 1),
    DEFINE_PAR("LeftRightScrollRepeat",leftright_button_repeat, PT_BOOL,   0, 1),
    DEFINE_PAR("ScrollButtonRepeat",   scroll_button_repeat,    PT_INT,    SBR_MIN , SBR_MAX),
    DEFINE_PAR("TouchpadOff",          touchpad_off,            PT_INT,    0, 2),
    DEFINE_PAR("LockedDrags",          locked_drags,            PT_BOOL,   0, 1),
    DEFINE_PAR("LockedDragTimeout",    locked_drag_time,        PT_INT,    0, 30000),
    DEFINE_PAR("RTCornerButton",       tap_action[RT_TAP],      PT_INT,    0, ELA_MAX_BUTTONS),
    DEFINE_PAR("RBCornerButton",       tap_action[RB_TAP],      PT_INT,    0, ELA_MAX_BUTTONS),
    DEFINE_PAR("LTCornerButton",       tap_action[LT_TAP],      PT_INT,    0, ELA_MAX_BUTTONS),
    DEFINE_PAR("LBCornerButton",       tap_action[LB_TAP],      PT_INT,    0, ELA_MAX_BUTTONS),
    DEFINE_PAR("TapButton1",           tap_action[F1_TAP],      PT_INT,    0, ELA_MAX_BUTTONS),
//    DEFINE_PAR("DragLock",           drag_lock,      PT_INT,    0, 1),
    DEFINE_PAR("TapButton2",           tap_action[F2_TAP],      PT_INT,    0, ELA_MAX_BUTTONS),
    DEFINE_PAR("TapButton3",           tap_action[F3_TAP],      PT_INT,    0, ELA_MAX_BUTTONS),
    DEFINE_PAR("GrabEventDevice",      grab_event_device,       PT_BOOL,   0, 1),
    DEFINE_PAR("ByteMode",             byte_mode,               PT_INT,    0, 10),
    DEFINE_PAR("DragTimeoutRelease",   drag_timeout_release,    PT_BOOL,   0, 1),
    DEFINE_PAR("PalmTrackingSlider",   palmtracking_slider,     PT_INT,    0, 15),
    DEFINE_PAR("DetectCovergestureIntr",detect_cover_intr,       PT_BOOL,    0, 1),
    DEFINE_PAR("DetectGestureIntr",    detect_gesture_intr,     PT_BOOL,    0, 1),
    DEFINE_PAR("DetectKbdIntr",        detect_kbd_intr,         PT_BOOL,    0, 1),
    DEFINE_PAR("DetectPalmtrackIntr",  detect_palmtrack_intr,   PT_BOOL,    0, 1),
    DEFINE_PAR("DetectTapIntr",        detect_tap_intr,         PT_BOOL,    0, 1),
    DEFINE_PAR("DetectEF_023TapIntr", detect_ef023_tap_intr,         PT_BOOL,    0, 1),
    { 0, 0, 0, 0, 0 }
};



static void
show_settings(ElantechSHM *elashm)
{
    int i;
     //ElantechSHM cur = *elashm;

    printf("Parameter settings:\n");
    for (i = 0; params[i].name; i++) {
        struct Parameter* par = &params[i];
        switch (par->type) {
        case PT_INT:
	    printf("    %-23s = %d\n", par->name, *(int*)((char*)elashm + par->offset));
            break;
        case PT_BOOL:
            printf("    %-23s = %d\n", par->name, *(Bool*)((char*)elashm + par->offset));
            break;
        case PT_DOUBLE:
            printf("    %-23s = %g\n", par->name, *(double*)((char*)elashm + par->offset));
            break;
        }
    }
}

static void
set_variables(ElantechSHM *elashm, int argc, char *argv[], int first_cmd)
{
    int i;
    for (i = first_cmd; i < argc; i++) {
        char *cmd = argv[i];
        char *eqp = index(cmd, '=');
        if (eqp) {
            int j;
            int found = 0;
            *eqp = 0;
            for (j = 0; params[j].name; j++) {
                if (strcasecmp(cmd, params[j].name) == 0) {
                    found = 1;
                    break;
                }
            }
            if (found) {
                double val = atof(&eqp[1]);
                struct Parameter* par = &params[j];

                if (val < par->min_val)
                    val = par->min_val;
                if (val > par->max_val)
                    val = par->max_val;

                switch (par->type) {
                case PT_INT:
                    *(int*)((char*)elashm + par->offset) = (int)rint(val);
                    break;
                case PT_BOOL:
                    *(Bool*)((char*)elashm + par->offset) = (Bool)rint(val);
                    break;
                case PT_DOUBLE:
                    *(double*)((char*)elashm + par->offset) = val;
                    break;
                }
            } else {
                printf("Unknown parameter %s\n", cmd);
            }
        } else {
            printf("Invalid command: %s\n", cmd);
        }
    }
}

static int
is_equal(ElantechSHM *s1, ElantechSHM *s2)
{
    int i;

    if ((s1->x           != s2->x) ||
        (s1->y           != s2->y) ||
        (s1->z           != s2->z) ||
	(s1->w		 != s2->w) ||
        (s1->x2_1        != s2->x2_1) ||
        (s1->y2_1        != s2->y2_1) ||
        (s1->x2_2        != s2->x2_2) ||
        (s1->y2_2        != s2->y2_2) ||
        (s1->x3_1        != s2->x3_1) ||
        (s1->y3_1        != s2->y3_1) ||
        (s1->numFingers  != s2->numFingers) ||
        (s1->left        != s2->left) ||
        (s1->right       != s2->right) ||
        (s1->up          != s2->up) ||
        (s1->down        != s2->down) ||
        (s1->middle      != s2->middle))
        return 0;
#if 0
    for (i = 0; i < 8; i++)
        if (s1->multi[i] != s2->multi[i])
            return 0;
#endif

    return 1;
}

static double
get_time()
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return tv.tv_sec + tv.tv_usec / 1000000.0;
}

static void
monitor(ElantechSHM *elashm, int delay)
{
    int header = 0;
    ElantechSHM old;
    double t0 = get_time();

    memset(&old, 0, sizeof(ElantechSHM));
    old.x = -1;                             /* Force first equality test to fail */

    while (1) {
        ElantechSHM cur = *elashm;
        if (!is_equal(&old, &cur) && cur.new_struct == 0) {
		int x,y,z,w,x2_1,y2_1,x2_2,y2_2,x3_1,y3_1;
		    x=y=z=w=x2_1=y2_1=x2_2=y2_2=x3_1=y3_1=0;
	    
	   switch(cur.numFingers){
		case 1:
			x=cur.x;
			y=cur.y;
			break;
		case 2:
			x2_1=cur.x2_1;
			y2_1=cur.y2_1;
			x2_2=cur.x2_2;
			y2_2=cur.y2_2;
			break;
		case 3:
			x3_1=cur.x3_1;
			y3_1=cur.y3_1;
			break;
		default:
			break;
	   }
		
	    z = cur.z;
	    w = cur.w;
		
		

            if (!header) {
                printf("%8s  %4s %4s %3s %3s %2s %2s %s %s %s %s  "
                       "%8s  %4s %4s  %4s %4s  %4s %4s\n",
                       "time", "x", "y", "z","w", "f", "l", "r", "u", "d", "m",
                       "x2_1","y2_1","x2_2","y2_2","x3_1","y3_1");
                header = 20;
            }
            header--;
            printf("%8.3f  %4d %4d %3d %3d %2d %2d %d %d %d %d "
                   "%4d %4d  %4d %4d  %4d %4d\n",
                   get_time() - t0,
                   x, y, z, w, cur.numFingers,
                   cur.left, cur.right, cur.up, cur.down, cur.middle,
                   x2_1, y2_1,x2_2, y2_2,x3_1,y3_1);
            fflush(stdout);
            old = cur;
        }
        usleep(delay * 1000);
    }
}

static void
usage()
{
    fprintf(stderr, "Usage: elaclient [-m interval] [-h] [-l] [-V] [-?] [var1=value1 [var2=value2] ...]\n");
    fprintf(stderr, "  -m monitor changes to the touchpad state.\n"
            "     interval specifies how often (in ms) to poll the touchpad state\n");
    fprintf(stderr, "  -h Show detected hardware properties\n");
    fprintf(stderr, "  -l List current user settings\n");
    fprintf(stderr, "  -V Print elaclient version string and exit\n");
    fprintf(stderr, "  -? Show this help message\n");
    fprintf(stderr, "  var=value  Set user parameter 'var' to 'value'.\n");
    exit(1);
}

int
main(int argc, char *argv[])
{
    ElantechSHM *elashm;
    int shmid;

    int c;
    int delay = -1;
    int do_monitor = 0;
    int dump_hw = 0;
    int dump_settings = 0;
    int first_cmd;
 
    /* Parse command line parameters */
    while ((c = getopt(argc, argv, "m:hlV")) != -1) {
        switch (c) {
        case 'm':
            do_monitor = 1;
            if ((delay = atoi(optarg)) < 0)
                usage();
            break;
        case 'l':
            dump_settings = 1;
            break;
        case 'V':
            printf("%s\n",ELAN_VERSION);
            exit(0);
        default:
            usage();
        }
    }
    first_cmd = optind;
    if (!do_monitor && !dump_hw && !dump_settings && first_cmd == argc)
        usage();

    /* Connect to the shared memory area */
    if ((shmid = shmget(SHM_ELANTECH, sizeof(ElantechSHM), 0)) == -1) {
        if ((shmid = shmget(SHM_ELANTECH, 0, 0)) == -1) {
            fprintf(stderr, "Can't access shared memory area. SHMConfig disabled?\n");
            exit(1);
        } else {
            fprintf(stderr, "Incorrect size of shared memory area. Incompatible driver version?\n");
            exit(1);
        }
    }
    if ((elashm = (ElantechSHM*) shmat(shmid, NULL, 0)) == NULL) {
        perror("shmat");
        exit(1);
    }

    /* Perform requested actions */
   
    set_variables(elashm, argc, argv, first_cmd);
    if (dump_settings) {
        show_settings(elashm);
    }
    if (do_monitor) {
        monitor(elashm, delay);
    }

    exit(0);
}
