/*   Elantech Passthrough Support
 *   Copyright (c) 2008 Elantech Inc Tom Lin<tom_lin@elantech.com.tw>
 *
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Trademarks are the property of their respective owners.
 *
 */

/*****************************************************************************
 *      Standard Headers
 ****************************************************************************/
#define xserver_version 1  


#if xserver_version

#include <unistd.h>
#include <sys/ioctl.h>
#include <misc.h>
#include <xf86.h>
#include <sys/shm.h>
#include <sys/ipc.h>
#include <sys/stat.h>
#include <errno.h>
#include <math.h>
#include <unistd.h>
#include <stdio.h>
#include <xf86_OSproc.h>
#include <xf86Xinput.h>
#include <exevents.h>
#include <xf86Optrec.h>
#include <xorg-server.h>
#include "mipointer.h"
#include <xorg-server.h>

#else
/* for xserver 1.3 */
#include <unistd.h>
#include <sys/ioctl.h>
#include <misc.h>
#include <xf86.h>
#include <sys/ipc.h>
#include <errno.h>
#include <math.h>
#include <unistd.h>
#define NEED_XF86_TYPES
#include <xf86_ansic.h>
#include <xf86_OSproc.h>
#include <xf86Xinput.h>
#include "mipointer.h"
#ifdef XFREE_4_0_3
#include <xf86Optrec.h>  		/* needed for Options */
#endif

#endif

/*****************************************************************************
 *      Local Headers
 ****************************************************************************/
#define ELANTECH_PRIVATE
#define ELAN_VERSION "ETD7.1.4 Free"
#include "elantech.h"

/*****************************************************************************
 *      Variables without includable headers
 ****************************************************************************/

/*****************************************************************************
 *      Local Variables and Types
 ****************************************************************************/
#define MAX(a, b) (((a)>(b))?(a):(b))
#define MIN(a, b) (((a)<(b))?(a):(b))
#define TIME_DIFF(a, b) ((int)((a)-(b)))
#define SYSCALL(call) while (((call) == -1) && (errno == EINTR))

#define SQR(x) ((x) * (x))
 
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 1
#define DBG(a,b)
#endif

/*****************************************************************************
 * Forward declaration
 ****************************************************************************/
 static InputInfoPtr ElantechPreInit(InputDriverPtr drv, IDevPtr dev,int flags);
 static void ElantechUnInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags);
 static Bool DeviceControl(DeviceIntPtr, int);
 static void ReadInput(LocalDevicePtr);
 static int HandleState(LocalDevicePtr,struct ElantechHwState*);
 static int ControlProc(LocalDevicePtr,xDeviceCtl*);
 static void CloseProc(LocalDevicePtr);
 static int SwitchMode(ClientPtr,DeviceIntPtr,int);
 static Bool ConvertProc(LocalDevicePtr, int, int, int, int, int, int, int, int, int*, int*);
 static Bool DeviceInit(DeviceIntPtr);
 static Bool DeviceOn(DeviceIntPtr);
 static Bool DeviceOff(DeviceIntPtr);
 static Bool DeviceClose(DeviceIntPtr);
 static Bool QueryHardware(LocalDevicePtr);
 static void ReadDevDimensions(LocalDevicePtr);
 
#if 0
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 3
void InitDeviceProperties(LocalDevicePtr local);
int SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
                BOOL checkonly);
#endif
#endif

InputDriverRec ELANTECH = {
    1,
    "elantech",
    NULL,
    ElantechPreInit,
    ElantechUnInit,
    NULL,
    0
};

static XF86ModuleVersionInfo VersionRec = {
    "elantech",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
#if xserver_version
    XORG_VERSION_CURRENT,
#else
    XF86_VERSION_CURRENT,
#endif
    1,0,0,
    ABI_CLASS_XINPUT,
    ABI_XINPUT_VERSION,
    MOD_CLASS_XINPUT,
    { 0, 0, 0, 0}
};
    
static pointer
SetupProc(pointer module, pointer options, int *errmaj, int *errmin)
{
    xf86AddInputDriver(&ELANTECH,module,0);
    return module;
}

XF86ModuleData elantechModuleData = {
    &VersionRec,
    &SetupProc,
    NULL
};

/*****************************************************************************
 *	Function Definitions
 ****************************************************************************/
static void
SetDeviceAndProtocol(LocalDevicePtr local)
{
    ElantechPrivate *priv = local->private;
    
    event_proto_operations.AutoDevProbe(local);
    priv->proto_ops = &event_proto_operations;
}

/*
 * Allocate and initialize memory for the SynapticsSHM data to hold driver
 * parameter settings.
 * The function will allocate shared memory if priv->shm_config is TRUE.
 * The allocated data is initialized from priv->synpara_default.
 */
static Bool
alloc_param_data(LocalDevicePtr local)
{
    int shmid;
    ElantechPrivate *priv = local->private;
    
    if(priv->elapara)
	return TRUE; /*Already allocated */


    if (priv->shm_config) {
#if xserver_version 
	if ((shmid = shmget (SHM_ELANTECH,0,0)) != -1)
	    shmctl(shmid,IPC_RMID,NULL);
	if ((shmid = shmget (SHM_ELANTECH,sizeof(ElantechSHM),
				0777 | IPC_CREAT)) == -1) {
#else
	if ((shmid = xf86shmget (SHM_ELANTECH,0,0)) != -1)
	    xf86shmctl(shmid,XF86IPC_RMID,NULL);
	if ((shmid = xf86shmget (SHM_ELANTECH,sizeof(ElantechSHM),
				0777 | XF86IPC_CREAT)) == -1) {
#endif
	    xf86Msg(X_ERROR, "%s error shmget\n",local->name);
	    return FALSE;
	}
#if xserver_version
	if ((priv->elapara = (ElantechSHM *)shmat(shmid, NULL, 0)) == NULL) {
#else
	if ((priv->elapara = (ElantechSHM *)xf86shmat(shmid, NULL, 0)) == NULL) {
#endif
	    xf86Msg(X_ERROR, "%s error shmget\n",local->name);
	    return FALSE;
	}
    } else {
	priv->elapara = xcalloc(1,sizeof(ElantechSHM));
	if(!priv->elapara)
	   return FALSE;
    }
    
    *(priv->elapara) = priv->elapara_default;
    return TRUE;
}
/*
 * Free SynapticsSHM data previously allocated by alloc_param_data().
 */
 static void
 free_param_data(ElantechPrivate *priv)
 {
    int shmid;
    
    if (!priv->elapara)
	return ;

    if (priv->shm_config) {
#if xserver_version
	  if ((shmid = shmget(SHM_ELANTECH,0,0)) != -1)
		shmctl (shmid,IPC_RMID,NULL);
#else
	  if ((shmid = xf86shmget(SHM_ELANTECH,0,0)) != -1)
		xf86shmctl (shmid,XF86IPC_RMID,NULL);
#endif
    } else {
	xfree(priv->elapara);
    }
    
    priv->elapara = NULL;
}

static double
elaSetFloatOption (pointer options,const char *optname, double default_value)
{
    char *str_par;
    double value;
    str_par = xf86FindOptionValue(options,optname);
#if xserver_version
    if ((!str_par) || (sscanf(str_par, "%lf",&value) != 1))
#else
    if ((!str_par) || (xf86sscanf(str_par, "%lf",&value) != 1))
#endif
	return default_value;
    return value;
}

#if 1
static void set_default_parameters(LocalDevicePtr local)
{

    ElantechPrivate *priv = local->private;	/*read-only */
    pointer opts = local->options;		/*read-only */
    ElantechSHM *pars = & priv->elapara_default; /*modified */
    
    int horizScrollDelta, vertScrollDelta;	/*pixels */
    int tapMove;	/* pixels */
    int l,r,t,b; /*left,right,top,botton */
    double accelFactor;				/*1/pixels*/
    int fingerLow,fingerHigh,fingerPress;	/*pressure*/
    /*int pressureMotionMinZ,pressureMotionMaxZ;*/
    int palmMinWidth,palmMinZ;			/*pressure*/
    int tapButton1,tapButton2,tapButton3;
    int clickFinger1,clickFinger2,clickFinger3;
    Bool vertEdgeScroll,horizEdgeScroll;
    Bool vertTwoFingerScroll,horizTwoFingerScroll;
    
    if (priv->minx < priv->maxx && priv->miny < priv->maxy) {
	
	int width,height,diag;
	int ewidth,eheight;	/*edge width/height */
	
	width = abs (priv->maxx - priv->minx);
	height = abs (priv->maxy - priv->miny);
	diag = sqrt(width * width + height * height);
	ewidth = width * .04;
	eheight = height * .055;
	
	l = priv->minx + ewidth;
	r = priv->maxx - ewidth;
	t = priv->miny + eheight;
	b = priv->maxy - eheight;
	
	/* Agian, based on typical x/y range and defaults */
	horizScrollDelta = diag * .020;
	vertScrollDelta = diag * .020;
	tapMove = diag * .044;
	accelFactor = 50.0 / diag;
    } else {
	l = priv->minx;
	r = priv->maxx;
	t = priv->miny;
	b = priv->maxy;
	
	horizScrollDelta = 100;
	vertScrollDelta = 100;
	tapMove = 200;
	accelFactor = 0.010;
    }
    
    if (priv->minp < priv->maxp) {
      int range = priv->maxp - priv->minp;
      
      /* scaling based on defaults below and a pressure of 256 */
      fingerLow = priv->minp + range * .098;
      fingerHigh = priv->minp + range * .117;
      fingerPress = priv->minp + range * 1.000;
      palmMinZ = priv->minp + range * .781;    
    } else {
      fingerLow = 25;
      fingerHigh = 30;
      fingerPress = 256;
      palmMinZ = 200;
    }
    
    if (priv->minw < priv->maxw) {
	int range = priv->maxw - priv->minw;
	
	/* scaling based on defaults below and a tool width of 16 */
	palmMinWidth = priv->minw + range * .625;
    } else {
        palmMinWidth = 10;
    } 
    
    /* Enable tap if we don't defaults below and tool width of 16 */
    tapButton1 = 1;
    tapButton2 = 2;
    tapButton3 = 3;
    
    /* Enable multifinger-click if we don't have right/middle button,
       otherwise clickFinger is always button 1. */
    clickFinger1 = 1;
    clickFinger2 = 2;
    clickFinger3 = 3;
    
    /*Enable vert edge scroll if we can detect doubletap */
    vertEdgeScroll =  TRUE;
    horizEdgeScroll = FALSE;
    
    /* Enable twofiner scroll if we can detect doubletap */
    vertTwoFingerScroll = TRUE;
    horizTwoFingerScroll = FALSE;
    
    /*set the parameters */
    pars->grab_event_device = xf86SetBoolOption(opts, "GrabEventDevice", TRUE);
    pars->left_edge = xf86SetIntOption(opts,"LeftEdge",l);
    pars->right_edge = xf86SetIntOption(opts, "RightEdge",r);
    pars->top_edge = xf86SetIntOption(opts, "TopEdge",t);
    pars->bottom_edge = xf86SetIntOption(opts,"BottomEdge",b);
    
    pars->finger_low = xf86SetIntOption(opts,"FingerLow",fingerLow);
    pars->finger_high = xf86SetIntOption(opts,"FingerHigh",fingerHigh);
    pars->finger_press = xf86SetIntOption(opts,"FingerPress",fingerPress);
    pars->tap_time = xf86SetIntOption(opts,"MaxTapTime",150);
    pars->tap_move = xf86SetIntOption(opts,"MaxTapMove",10);
    pars->tap_time_2 = xf86SetIntOption(opts,"MaxDoubleTapTime",180);
    pars->click_time = xf86SetIntOption(opts,"ClickTime",200);
    /*pars->fast_taps = xf86SetIntOption(opts, "FastTaps",FALSE);*/
    pars->emulate_mid_button_time = xf86SetIntOption(opts,"EmulateMidButtonTime",75);
   /*pars->emulate_twofinger_z = xf86SetIntOption(opts,"EmulateTwoFingerMinZ",emulateTwoFingerMinZ); tom -*/
    pars->scroll_dist_vert = xf86SetIntOption(opts,"VertScrollDelta",vertScrollDelta);
    pars->scroll_dist_horiz = xf86SetIntOption(opts,"HorizScrollDelta",horizScrollDelta);
    pars->updown_button_scrolling = xf86SetBoolOption(opts, "UpDownScrolling", TRUE);
    pars->leftright_button_scrolling = xf86SetBoolOption(opts, "LeftRightScrolling", TRUE);
    pars->updown_button_repeat = xf86SetBoolOption(opts, "UpDownScrollRepeat", TRUE);
    pars->leftright_button_repeat = xf86SetBoolOption(opts, "LeftRightScrollRepeat", TRUE);
    pars->scroll_button_repeat = xf86SetIntOption(opts,"ScrollButtonRepeat", 100);
    
    pars->touchpad_off = xf86SetIntOption(opts, "TouchpadOff",0);
    pars->locked_drags = xf86SetBoolOption(opts, "LockedDrags",FALSE);
    pars->locked_drag_time = xf86SetIntOption(opts, "LockedDragTimeout", 5000);
    pars->single_tap_timeout = xf86SetIntOption(opts,"SingleTapTimeout",180);
    pars->tap_action[RT_TAP] = xf86SetIntOption(opts,"RTCornerButton",0);
    pars->tap_action[RB_TAP] = xf86SetIntOption(opts,"RBCornerButton",0);
    pars->tap_action[LT_TAP] = xf86SetIntOption(opts,"LTCornerButton",0);
    pars->tap_action[LB_TAP] = xf86SetIntOption(opts,"LBCornerButton",0);
    pars->tap_action[F1_TAP] = xf86SetIntOption(opts,"TapButton1",tapButton1);
    pars->tap_action[F2_TAP] = xf86SetIntOption(opts,"TapButton2",tapButton2);
    pars->tap_action[F3_TAP] = xf86SetIntOption(opts,"TapButton3",tapButton3);
  
    pars->min_speed = elaSetFloatOption(opts,"MinSpeed",0.4);
    pars->max_speed = elaSetFloatOption(opts,"MaxSpeed",0.7);
    pars->accl = elaSetFloatOption (opts,"AccelFactor",accelFactor);
    pars->trackstick_speed = elaSetFloatOption (opts,"TrackstickSpeed",40);
   
      /* Warn about (and fix) incorrectly configured TopEdge/BottomEdge parameters */
    if (pars->top_edge > pars->bottom_edge) {
	int tmp = pars->top_edge;
	pars->top_edge = pars->bottom_edge;
	pars->bottom_edge = tmp;
	xf86Msg(X_WARNING, "%s: TopEdge is bigger than BottomEdge. Fixing.\n",
		local->name);
    }
    
    
    pars->button_checkBox = xf86SetBoolOption(opts, "ButtonCheckBox",TRUE);
    pars->button_left_comboBox = xf86SetIntOption(opts, "ButtonLeftComboBox",0);
    pars->button_middle_comboBox = xf86SetIntOption(opts,"ButtonMiddleComboBox",2);
    pars->button_right_comboBox = xf86SetIntOption(opts,"ButtonRightComboBox",1);
    pars->tapping_checkBox =  xf86SetBoolOption(opts, "TappingCheckBox", TRUE);
    pars->two_FT_comboBox = xf86SetIntOption(opts,"TwoFTComboBox",7);
    pars->three_FT_comboBox = xf86SetIntOption(opts,"ThreeFTComboBox",7);
    pars->disable_when_key_checkBox = xf86SetBoolOption(opts, "DisableWhenKeyCheckBox", TRUE);
    pars->tapping_delay_comboBox = xf86SetIntOption(opts,"TappingDelayComboBox",15);
    pars->cursor_delay_comboBox = xf86SetIntOption(opts,"CursorDelayComboBox",4);
    pars->gesture_delay_comboBox = xf86SetIntOption(opts,"GestureDelayComboBox",15);
    pars->cover_delay_comboBox = xf86SetIntOption(opts,"CoverDelayComboBox",4);
    pars->palmtrack_delay_comboBox = xf86SetIntOption(opts,"PalmtrackDelayComboBox",0);
    pars->edge_v_scroll_checkBox = xf86SetBoolOption(opts, "EdgeVscrollCheckBox", FALSE);
    pars->edge_h_scroll_checkBox = xf86SetBoolOption(opts, "EdgeHscrollCheckBox", FALSE);
    pars->edge_scroll_speed_slider = xf86SetIntOption(opts,"EdgeScrollSpeedSlider",5);
    pars->edge_scroll_area_slider = xf86SetIntOption(opts,"EdgeScrollAreaSlider",1);
    pars->vertical_scrolling_checkBox = xf86SetBoolOption(opts, "VerticalScrollingCheckBox", TRUE);
    pars->horizontal_scrolling_checkBox = xf86SetBoolOption(opts, "HorizontalScrollingCheckBox", TRUE);
    pars->scrolling_speed_slider = xf86SetIntOption(opts,"ScrollingSpeedSlider",4); 
    pars->auto_scroll_checkBox = xf86SetBoolOption(opts, "AutoScrollCheckBox", TRUE);
    pars->zoom_checkBox = xf86SetBoolOption(opts, "ZoomCheckBox", TRUE);
    pars->zoom_out_comboBox = xf86SetIntOption(opts,"ZoomOutComboBox",2);
    pars->zoom_in_comboBox = xf86SetIntOption(opts,"ZoomInComboBox",3);
    pars->zoom_slider = xf86SetIntOption(opts,"ZoomSlider",9);
    pars->rotation_checkBox = xf86SetBoolOption(opts, "RotationCheckBox", TRUE);
    pars->clockwise_comboBox = xf86SetIntOption(opts,"ClockwiseComboBox",2);
    pars->counterclosewise_comboBox = xf86SetIntOption(opts,"CounterclosewiseComboBox",3);
    pars->rotation_slider = xf86SetIntOption(opts,"RotationSlider",7);
    pars->drag_checkBox = xf86SetBoolOption(opts, "DragCheckBox", TRUE);
    pars->lift_release_radioButton = xf86SetBoolOption(opts, "LiftReleaseRadioButton", TRUE);
    pars->tap_again_radioButton = xf86SetBoolOption(opts, "TapAgainRadioButton",FALSE);
    pars->tap_release_radioButton = xf86SetBoolOption(opts, "TapReleaseRadioButton", FALSE);
    pars->drag_slider = elaSetFloatOption(opts, "DragSlider", 1.5);
    pars->swipe_page_checkBox = xf86SetBoolOption(opts, "SwipePageCheckBox", FALSE);
    pars->fast_key_checkBox = xf86SetBoolOption(opts, "FastKeyCheckBox", FALSE);
    pars->my_home_file_checkBox = xf86SetBoolOption(opts, "MyHomeFileCheckBox", FALSE);
    pars->switch_windows_checkBox = xf86SetBoolOption(opts, "SwitchWindowsCheckBox", FALSE);
    pars->covergesture_checkBox = xf86SetBoolOption(opts, "CovergestureCheckBox", TRUE);
    pars->covergesture_slider = xf86SetIntOption(opts,"CovergestureSlider",9);
    pars->covergesture_comboBox = xf86SetIntOption(opts,"CovergestureComboBox",17);
    pars->palmtracking_checkBox = xf86SetBoolOption(opts, "PalmtrackingCheckBox", TRUE);
    
    pars->new_struct = xf86SetIntOption(opts,"NewStruct",0);
    pars->detect_kbd_intr = xf86SetBoolOption(opts, "DetectKbdIntr", FALSE);
    pars->detect_tap_intr = xf86SetBoolOption(opts, "DetectTapIntr", FALSE);
    pars->detect_gesture_intr = xf86SetBoolOption(opts, "DetectGestureIntr", FALSE);
    pars->detect_cover_intr = xf86SetBoolOption(opts, "DetectCoverIntr", FALSE);
    pars->detect_palmtrack_intr = xf86SetBoolOption(opts, "DetectPalmtrackIntr", FALSE);
    
    pars->byte_mode = priv->byte_mode;
    pars->version = priv->version;
    pars->has_pressure = priv->has_pressure;
    pars->smart_pad = priv->smart_pad;
    pars->convergesture = priv->convergesture;
}   
#endif
 /*
 *  called by the module loader for initialization
 */
 static InputInfoPtr
 ElantechPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
{
    LocalDevicePtr local;
    ElantechPrivate *priv;
    
    /* allocate memory for ElantechPrivateRec */
    priv = xcalloc(1,sizeof(ElantechPrivate));
    if(!priv)
	return NULL;

    /* Allocate a new InputInfoRec and add it to the head xf86InputDev. */
    local = xf86AllocateInput(drv,0);
    if (!local) {
	xfree(priv);
	return NULL;
    }
    
    /* initialize the InputInfoRec */
    local->name                     = dev->identifier;
#if xserver_version
    local->type_name                = XI_TOUCHPAD;
#else
    local->type_name                = XI_MOUSE;
#endif
    local->device_control           = DeviceControl;
    local->read_input               = ReadInput;
    local->control_proc             = ControlProc;
    local->close_proc               = CloseProc;
    local->switch_mode              = SwitchMode;
    local->conversion_proc          = ConvertProc;
    local->reverse_conversion_proc  = NULL;
    local->dev                      = NULL;
    local->private                  = priv;
    local->private_flags            = 0;
    local->flags                    = XI86_POINTER_CAPABLE | XI86_SEND_DRAG_EVENTS;
    local->conf_idev                = dev;
#if xserver_version    
  #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
    local->motion_history_proc      = xf86GetMotionEvents;
    local->history_size             = 0;
  #endif
#else
    local->motion_history_proc      = xf86GetMotionEvents;
    local->history_size             = 0;
#endif
    local->always_core_feedback     = 0;
    
    xf86Msg(X_INFO,"Elantech touchpad driver version %s",ELAN_VERSION);
    
    xf86CollectInputOptions(local,NULL,NULL);
    
    xf86OptionListReport(local->options);
    
    /* may change local->options */
    SetDeviceAndProtocol(local);
    
    /* open the touchpad device */
    local->fd = xf86OpenSerial(local->options);
    if(local->fd == -1) {
       xf86Msg(X_ERROR,"Elantech driver unable to open driver\n");
       goto SetupProc_fail;
    }
    
    xf86ErrorFVerb(6,"port opened successfully\n");
    
    /* initialize variables */
    priv->timer = NULL;
    priv->repeatButtons = 0;
    priv->nextRepeat = 0;
    priv->count_packet_finger = 0;
    priv->tap_state = TS_START;
    priv->tap_button = 0;
    priv->tap_button_state = TBS_BUTTON_UP;
    priv->touch_on.millis = 0;
    
    /*read hardware dimensions */
    ReadDevDimensions(local);
    
    /* install shared memory or normal memory for parameters */
    priv->shm_config = xf86SetBoolOption(local->options,"SHMConfig",FALSE);
    
    set_default_parameters(local);
    
   /*priv->largest_valid_x = MIN(priv->elapara_default.right_edge,XMAX_NOMINAL); tom -*/
   
    
    if(!alloc_param_data(local))
	goto SetupProc_fail;
	
    priv->comm.buffer = XisbNew(local->fd,200);
    DBG(9,XisbTrace(priv->comm.buffer,1));
    
    if (!QueryHardware(local)) {
        xf86Msg(X_ERROR, "%s Unable to query/initialize Elantech hardware.\n",local->name);
	goto SetupProc_fail;
    }
    
    xf86ProcessCommonOptions(local,local->options);
    local->flags |= XI86_CONFIGURED;
    
    if (local->fd != 1) {
	if (priv->comm.buffer) {
	    XisbFree(priv->comm.buffer);
	    priv->comm.buffer = NULL;
	}
	xf86CloseSerial(local->fd);
    }
    local->fd = -1;
    
    return local;
 
SetupProc_fail:
    if(local->fd >=0) {
	xf86CloseSerial(local->fd);
	local->fd = -1;
    }
    
    if (priv->comm.buffer)
	XisbFree(priv->comm.buffer);
    free_param_data(priv);
    xfree(priv);
    local->private = NULL;
    
     return local;
}


/*
 *  Uninitialize the device.
 */
 static void ElantechUnInit(InputDriverPtr drv,
			    InputInfoPtr   local,
			    int            flags)
{
    xfree(local->private);
    local->private = NULL;
    xf86DeleteInput(local,0);
}

/*
 *  Alter the control parameters for the mouse. Note that all special
 *  protocol values are handled by dix.
 */
static void 
ElantechCtrl(DeviceIntPtr device, PtrCtrl *ctrl)
{
      DBG(3, ErrorF("ElantechCtrl called.\n"));
    /*
      pInfo = device->public.devicePrivate;
      pMse = pInfo->private;

      pMse->num       = ctrl->num;
      pMse->den       = ctrl->den;
      pMse->threshold = ctrl->threshold;
    */
}
static Bool
DeviceControl(DeviceIntPtr dev, int mode)
{
    Bool RetValue;
    
    switch (mode) {
	case DEVICE_INIT:
	    RetValue = DeviceInit(dev);
	    break;
	case DEVICE_ON:
	    RetValue = DeviceOn(dev);
	    break;
	case DEVICE_OFF:
	    RetValue = DeviceOff(dev);
	    break;
	case DEVICE_CLOSE:
	    RetValue = DeviceClose(dev);
	    break;
	default:
	    RetValue = BadValue;
   }
   
   return RetValue;
}
    
static Bool
DeviceInit(DeviceIntPtr dev)
{
    LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
    ElantechPrivate *priv = (ElantechPrivate *) (local->private);
    unsigned char map[ELA_MAX_BUTTONS +1];
    int i;
    
    DBG(3,ErrorF("Elantech DeviceInit called\n"));
    
    for (i = 0; i <= ELA_MAX_BUTTONS;i++)
	map[i] = i;
	
    dev->public.on = FALSE;

#if xserver_version    
    InitPointerDeviceStruct((DevicePtr)dev, map,
			      ELA_MAX_BUTTONS,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
			      miPointerGetMotionEvents,
#elif GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 3
			      GetMotionHistory,
#endif
			      ElantechCtrl,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
			      miPointerGetMotionBufferSize()
#else
			      GetMotionHistorySize(),2
#endif
			     );
#else
    InitPointerDeviceStruct((DevicePtr)dev, map,
			    ELA_MAX_BUTTONS,
			    miPointerGetMotionEvents, ElantechCtrl,
			    miPointerGetMotionBufferSize());
#endif
			     
	/* X valuator */
	if (priv->minx < priv->maxx)
	    xf86InitValuatorAxisStruct(dev, 0, priv->minx, priv->maxx, 1, 0, 1);
	else
	    xf86InitValuatorAxisStruct(dev, 0, 0, -1, 1, 0, 1);
	xf86InitValuatorDefaults(dev,0);
	
	/* Y valuator */
	if (priv->miny < priv->maxy)
	    xf86InitValuatorAxisStruct(dev,1,priv->miny,priv->maxy,1,0,1);
	else
	    xf86InitValuatorAxisStruct(dev, 1, 0, -1, 1, 0, 1);
	xf86InitValuatorDefaults(dev,1);

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
	xf86MotionHistoryAllocate(local);
#endif
	if (!alloc_param_data(local))
	      return !Success;
#if 0
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 3
	InitDeviceProperties(local);
	XIRegisterPropertyHandler(local->dev,SetProperty,NULL,NULL);
#endif
#endif
	return Success;
	
}

static Bool
DeviceOn(DeviceIntPtr dev)
{
    LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
    ElantechPrivate *priv = (ElantechPrivate *)(local->private);
    
    DBG(3,ErrorF("Elantech DeviceOn called\n"));
    
    SetDeviceAndProtocol(local);
    local->fd = xf86OpenSerial(local->options);
    if (local->fd == -1) {
	xf86Msg(X_WARNING, "%s: cannot open inpput device\n",local->name);
	return !Success;
    }
    
    priv->proto_ops->DeviceOnHook(local,priv->elapara);
    
    priv->comm.buffer = XisbNew(local->fd ,64);
    if (!priv->comm.buffer) {
	  xf86CloseSerial(local->fd);
	  local->fd = -1;
	  return !Success;
    }
    
    xf86FlushInput(local->fd);
    
    /* reinit the pad */
    QueryHardware(local);
    xf86AddEnabledDevice(local);
    dev->public.on = TRUE;
    
    return Success;
}

static Bool
DeviceOff(DeviceIntPtr dev)
{
    LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
    ElantechPrivate *priv = (ElantechPrivate *) (local->private);
    
    DBG(3,ErrorF("Elantech DeviceOff called\n"));
    
    if (local->fd != -1) {
	TimerFree(priv->timer);
	priv->timer = NULL;
	xf86RemoveEnabledDevice(local);
	priv->proto_ops->DeviceOffHook(local);
	if (priv->comm.buffer) {
	    XisbFree(priv->comm.buffer);
	    priv->comm.buffer = NULL;
	}
	xf86CloseSerial(local->fd);
    }
    dev->public.on = FALSE;
    return Success;
}

static Bool
DeviceClose(DeviceIntPtr dev)
{
    Bool RetValue;
    LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
    ElantechPrivate *priv = (ElantechPrivate *) local->private;
    
    RetValue = DeviceOff(dev);
    free_param_data(priv);
    return RetValue;
}

struct ScrollData 
{
    int left, right, up, down;
};

static void
HandleScrollProcessing(ElantechPrivate *priv,int x, int y, struct ScrollData *sd)
{
	ElantechSHM *para = priv->elapara;
	int dx, dy;
	static int one_finger_x,one_finger_y;
	static int county_forward;
	static int countx_forward;
	static int countx_back;
	static int county_back;
	int NowDeltaX21;
	int NowDeltaY21;
	int PreDeltaX21;
	int PreDeltaY21;
	int DeltaX1;
	int DeltaY1;
	int DeltaX2;
	int DeltaY2;
	static int last_x1;
	static int last_y1;
	static int last_x2;
	static int last_y2;
	static int can_edge_scroll_H_function=0;
	static int can_edge_scroll_V_function=0;

	static int TwoFingerScrollUpCounter=0;
	static int TwoFingerScrollDownCounter=0;
	static int TwoFingerScrollRightCounter=0;
	static int TwoFingerScrollLeftCounter = 0;
  
	static int ScrollValue_Up=0;
	static int ScrollValue_Down=0;
	static int ScrollValue_Left=0;
	static int ScrollValue_Right=0;

	static int dont_known_edge=0;

	int ScrollDiv_LeftRight =0;
	int ScrollDiv_UpDown=0;
	static char	PreScrollFlag	= 0x00;
	char	ScrollFlag = 0x00;
	int scroll_temp=16;
	int STV_Up_Down;
	int STV_Left_Right;
	
	sd->left = sd->right = sd->up = sd->down = 0;
	
	dx = dy = 0;
	/*two finger scroll */
	DeltaX1 = (para->x2_1 - last_x1);
	DeltaY1 = (para->y2_1 - last_y1);
	DeltaX2 = (para->x2_2 - last_x2);
	DeltaY2 = (para->y2_2 - last_y2);

	NowDeltaX21=(para->x2_2-para->x2_1);
	NowDeltaY21=(para->y2_2-para->y2_1);

	PreDeltaX21=(last_x2-last_x1);
	PreDeltaY21=(last_y2-last_y1);
	

	if((para->numFingers==2) && (para->new_struct == 0))
	{
		
	     	if(( PreDeltaY21-NowDeltaY21 ) < 3  && abs(DeltaY1-DeltaY2) < 3 && para->vertical_scrolling_checkBox==1)
		{
			if(DeltaY1 < 0 && DeltaY2 < 0)
			{
				if(DeltaY1 > DeltaY2)
				{
					TwoFingerScrollUpCounter++;
					ScrollValue_Up += abs(DeltaY2); 
					ScrollValue_Down = 0;
					TwoFingerScrollDownCounter = 0;
				}
				else
				{
					TwoFingerScrollUpCounter++;
					ScrollValue_Up += abs(DeltaY1); 
					ScrollValue_Down = 0;
					TwoFingerScrollDownCounter = 0;
				}
			}
			else if(DeltaY1 > 0 && DeltaY2 > 0)
			{
				if(DeltaY1 > DeltaY2)
				{
					TwoFingerScrollDownCounter++;
					ScrollValue_Up = 0;
					ScrollValue_Down += abs(DeltaY2); 
					TwoFingerScrollUpCounter = 0;
				}
				else
				{
					TwoFingerScrollDownCounter++;
					ScrollValue_Up = 0;
					ScrollValue_Down += abs(DeltaY1); 
					TwoFingerScrollUpCounter = 0;
				}
			}
			STV_Up_Down = 0;
			/*Up*/
			if((PreScrollFlag & 0x01) == 0x01 && TwoFingerScrollUpCounter == 1)
			{
				STV_Up_Down = 2;
				ScrollValue_Up = 0;
			}
			else
			{
				STV_Up_Down = 1;
			}
			/*Down*/
			if((PreScrollFlag & 0x02) == 0x02 && TwoFingerScrollDownCounter == 1)
			{
				STV_Up_Down = 2;
				ScrollValue_Down = 0;
			}
			else
			{
				STV_Up_Down = 1;
			}
						
			/*----------------------------------------------------------*/
			if(TwoFingerScrollUpCounter >= STV_Up_Down)
			{
				ScrollDiv_UpDown = ScrollValue_Up / (scroll_temp - para->scrolling_speed_slider);
				if(ScrollDiv_UpDown > 0)
				{
					ScrollFlag |= 0x02;
				}		
			}
							
			else if(TwoFingerScrollDownCounter >= STV_Up_Down)
			{
				ScrollDiv_UpDown = ScrollValue_Down / (scroll_temp - para->scrolling_speed_slider);
				if(ScrollDiv_UpDown > 0)
				{
					ScrollFlag |= 0x01;
				}
			}
					
		}
			
		if( ( NowDeltaX21-PreDeltaX21 ) < 3 && abs(DeltaX1-DeltaX2) < 3 && para->horizontal_scrolling_checkBox ==1)
		{
			if(DeltaX1 > 0 && DeltaX2 > 0)
			{
				if(DeltaX1 > DeltaX2)
				{
					TwoFingerScrollRightCounter++;
					ScrollValue_Left = 0;
					ScrollValue_Right += abs(DeltaX2); 
					TwoFingerScrollLeftCounter = 0;
				}
				else
				{
					TwoFingerScrollRightCounter++;
					ScrollValue_Left = 0;
					ScrollValue_Right += abs(DeltaX1); 
					TwoFingerScrollLeftCounter = 0;
				}
			}
			else if(DeltaX1 < 0 && DeltaX2 < 0)
			{
				if(DeltaX1 > DeltaX2)
				{
					TwoFingerScrollLeftCounter++;
					ScrollValue_Left += abs(DeltaX2); 
					ScrollValue_Right = 0;
					TwoFingerScrollRightCounter = 0;
				}
				else
				{
					TwoFingerScrollLeftCounter++;
					ScrollValue_Left += abs(DeltaX1); 
					ScrollValue_Right = 0;
					TwoFingerScrollRightCounter = 0;
				}
			}
			STV_Left_Right = 0;
			/*Right*/
			if((PreScrollFlag & 0x10) == 0x10 &&  TwoFingerScrollRightCounter == 1)
			{
				STV_Left_Right = 2;
				ScrollValue_Right = 0;
			}
			else
			{
				STV_Left_Right = 1;
			}
			/*Left*/
			if((PreScrollFlag & 0x20) == 0x20 && TwoFingerScrollLeftCounter == 1)
			{
				STV_Left_Right = 2;
				ScrollValue_Left = 0;
			}
			else
			{
				STV_Left_Right = 1;
			}


			if(TwoFingerScrollRightCounter >= STV_Left_Right)
			{
				ScrollDiv_LeftRight = ScrollValue_Right / (scroll_temp - para->scrolling_speed_slider);
				
				if(ScrollDiv_LeftRight > 0)
				{
					ScrollFlag |= 0x20;
				}
			}
			else if(TwoFingerScrollLeftCounter >= STV_Left_Right)
			{
				ScrollDiv_LeftRight = ScrollValue_Left / (scroll_temp - para->scrolling_speed_slider);
				if(ScrollDiv_LeftRight > 0)
				{
					ScrollFlag |= 0x10;
				}
			}
		}
	
		PreScrollFlag = ScrollFlag;
		if((ScrollFlag & 0x01) == 0x01)
		{
			/*printf("Scroll_Down_fuction\n");*/
			sd->down=1;
			
			ScrollValue_Up = 0;
			ScrollValue_Down = 0;
			ScrollDiv_UpDown = 0;
			TwoFingerScrollUpCounter	= 0;
			TwoFingerScrollDownCounter	= 0;
			TwoFingerScrollRightCounter 	= 0;
			TwoFingerScrollLeftCounter 	= 0;
		}
		else if((ScrollFlag & 0x02) == 0x02)
		{
			/*printf("Scroll_Up_fuction\n");*/
			sd->up =1;
				
			ScrollValue_Up = 0;
			ScrollValue_Down = 0;
			ScrollDiv_UpDown = 0;
			TwoFingerScrollUpCounter	= 0;
			TwoFingerScrollDownCounter	= 0;
			TwoFingerScrollRightCounter 	= 0;
			TwoFingerScrollLeftCounter 	= 0;
		}
		else if((ScrollFlag & 0x10) == 0x10)
		{
			/*printf("Scroll_Left_fuction\n");*/			
			sd->left =1;
			ScrollValue_Right = 0;
			ScrollValue_Left = 0;
			ScrollDiv_LeftRight = 0;
			TwoFingerScrollUpCounter	= 0;
			TwoFingerScrollDownCounter	= 0;
			TwoFingerScrollRightCounter 	= 0;
			TwoFingerScrollLeftCounter 	= 0;
		}
		else if((ScrollFlag & 0x20) == 0x20)
		{
			/*printf("Scroll_Right_fuction\n");*/
			sd->right =1;
			ScrollValue_Right = 0;
			ScrollValue_Left = 0;
			ScrollDiv_LeftRight = 0;
			TwoFingerScrollUpCounter	= 0;
			TwoFingerScrollDownCounter	= 0;
			TwoFingerScrollRightCounter 	= 0;
			TwoFingerScrollLeftCounter 	= 0;
		}
		last_x1 = para->x2_1;
		last_y1 = para->y2_1;
		last_x2 = para->x2_2;
		last_y2 = para->y2_2;
	}
	
        /*edge scroll*/
	
	if((para->numFingers == 1) && para->all_can_move==0)
	{
		if((para->x > (priv->maxx * 0.6 + (11-para->edge_scroll_area_slider)*(priv->maxx - priv->maxx*0.6)/11)) && 
		   (para->y > (priv->maxy * 0.6 + 11-para->edge_scroll_area_slider*(priv->maxy-priv->maxy* 0.6)/11)) && 
		    dont_known_edge==0 && (para->edge_v_scroll_checkBox==1 && para->edge_h_scroll_checkBox==1))
		{
			    para->can_edge_scroll_function=1;
			    dont_known_edge=0;
		}
		
		else if((para->y >  (priv->maxy * 0.6 + (11-para->edge_scroll_area_slider)*(priv->maxy - priv->maxy * 0.6)/11 )) && 
			 can_edge_scroll_V_function==0  )
		{	
			if(para->edge_h_scroll_checkBox==1)
			{
			    dx = (para->x) - one_finger_x;
			    one_finger_x = para->x;
			    dont_known_edge=1;
			    para->can_edge_scroll_function=1;
			    can_edge_scroll_H_function=1;
			    
			    if(dx > 0 )
			    {
			      countx_forward++;
			    }
			    else
			    {
			      countx_back++;
			    }
			}
		}
		else if((para->x > (priv->maxx*0.6+(11-para->edge_scroll_area_slider)*(priv->maxx-priv->maxx*0.6)/11)) && 
			 can_edge_scroll_H_function==0 )
		{
			if(para->edge_v_scroll_checkBox==1)
			{
			      dy = (para->y) - one_finger_y;
			      one_finger_y = para->y;			
			      dont_known_edge=1;
			      can_edge_scroll_V_function=1;
			
			      if(dy > 0 )
			      {
		      		county_forward++;
			      }
			      else
			      {
	      			county_back++;
			      }
			      para->can_edge_scroll_function=1;
			}
		}
		else
		{
			dont_known_edge=0;
			para->all_can_move=1;
			para->can_edge_scroll_function=0;
			can_edge_scroll_V_function=0;
			can_edge_scroll_H_function=0;
		}
	}

	if(para->numFingers == 0)
	{
		para->all_can_move=0;
		countx_forward=0;
		county_back=0;
		countx_back=0;
		county_forward=0;
		para->can_edge_scroll_function=0;
		dont_known_edge=0;
		can_edge_scroll_V_function=0;
		can_edge_scroll_H_function=0;
		PreScrollFlag=0x00;
		
	}

	if(para->can_edge_scroll_function==1 && dont_known_edge==1)
	{
		if(dx < 0 && countx_back>(12-para->edge_scroll_speed_slider))
		{
			countx_forward=0;
			county_back=0;
			countx_back=0;
			county_forward=0;
               		sd->left  = 1;
		}
        	else if(dx > 0 && countx_forward>(12-para->edge_scroll_speed_slider))
		{ 
			countx_forward=0;
			county_back=0;
			countx_back=0;
			county_forward=0;
                       sd->right = 1;
		}
        	if(dy < 0 && county_back>(12-para->edge_scroll_speed_slider))
		{
			countx_forward=0;
			county_back=0;
			countx_back=0;
			county_forward=0;
                       sd->up    = 1;
		}
        	else if(dy >0 && county_forward>(12-para->edge_scroll_speed_slider))
		{
			countx_forward=0;
			county_back=0;
			countx_back=0;
			county_forward=0;
                        sd->down  = 1;
		}
	}
}


static CARD32
timerFunc(OsTimerPtr timer, CARD32 now, pointer arg)
{
    LocalDevicePtr local = (LocalDevicePtr) (arg);
    ElantechPrivate *priv = (ElantechPrivate *) (local->private);
    struct ElantechHwState hw;
    int delay;
    int sigstate;
    CARD32 wakeUpTime;
    
    sigstate = xf86BlockSIGIO();
    
    hw = priv->hwState;
   
    hw.millis = now;
    delay = HandleState(local, &hw);
     /*
     * Workaround for wraparound bug in the TimerSet function. This bug is already
     * fixed in CVS, but this driver needs to work with XFree86 versions 4.2.x and
     * 4.3.x too.
     */
    wakeUpTime = now + delay;
    if (wakeUpTime <= now)
	wakeUpTime = 0xffffffffL;

    priv->timer = TimerSet(priv->timer, TimerAbsolute, wakeUpTime, timerFunc, local);

    xf86UnblockSIGIO(sigstate);

    return 0;
}

static Bool
ElantechGetHwState(LocalDevicePtr local, ElantechPrivate *priv,
		   struct ElantechHwState *hw)
{
     return priv->proto_ops->ReadHwState(local,priv->proto_ops,
					  &priv->comm,hw);
}

    
/*
 *  called for each full received packet from the touchpad
 */
 static void
 ReadInput(LocalDevicePtr local)
 {
      ElantechPrivate *priv = (ElantechPrivate *) (local->private);
      struct ElantechHwState hw;
      int delay = 0;
      Bool newDelay = FALSE;
      
      while (ElantechGetHwState(local,priv,&hw)){
	 hw.millis = GetTimeInMillis();
	 priv->hwState = hw;
	 delay = HandleState(local,&hw);
	 newDelay = TRUE;
      }
      if(newDelay)
	 priv->timer = TimerSet(priv->timer, 0, delay, timerFunc, local);
}  

static int
HandleMidButtonEmulation(ElantechPrivate *priv, struct ElantechHwState *hw, int *delay)
{
    ElantechSHM *para = priv->elapara;
    Bool done = FALSE;
    int timeleft;
    int mid = 0;

    while (!done) {
        switch (priv->mid_emu_state) {
        case MBE_OFF:
            priv->button_delay_millis = hw->millis;
            if (para->left) {
                priv->mid_emu_state = MBE_LEFT;
            } else if (para->right) {
                priv->mid_emu_state = MBE_RIGHT;
            } else {
                done = TRUE;
            }
            break;
        case MBE_LEFT:
            timeleft = TIME_DIFF(priv->button_delay_millis + para->emulate_mid_button_time,
                                 hw->millis);
            if (timeleft > 0)
                *delay = MIN(*delay, timeleft);
            if (!para->left || (timeleft <= 0)) {
                para->left = TRUE;
                priv->mid_emu_state = MBE_TIMEOUT;
                done = TRUE;
            } else if (para->right) {
                priv->mid_emu_state = MBE_MID;
            } else {
                para->left = FALSE;
                done = TRUE;
            }
            break;
        case MBE_RIGHT:
            timeleft = TIME_DIFF(priv->button_delay_millis + para->emulate_mid_button_time,
                                 hw->millis);
            if (timeleft > 0)
                *delay = MIN(*delay, timeleft);
            if (!para->right || (timeleft <= 0)) {
                para->right = TRUE;
                priv->mid_emu_state = MBE_TIMEOUT;
                done = TRUE;
            } else if (para->left) {
                priv->mid_emu_state = MBE_MID;
            } else {
                para->right = FALSE;
                done = TRUE;
            }
            break;
        case MBE_MID:
            if (!para->left && !para->right) {
                priv->mid_emu_state = MBE_OFF;
            } else {
                mid = TRUE;
                para->left = para->right = FALSE;
                done = TRUE;
            }
            break;
        case MBE_TIMEOUT:
            if (!para->left && !para->right) {
                priv->mid_emu_state = MBE_OFF;
            } else {
                done = TRUE;
            }
        }
    }
    return mid;
}



static int
ElantechDetectFinger(ElantechPrivate *priv, struct ElantechHwState *hw)
{
       
        int finger;
        
        /*finger detection thru pressure and threshold*/
        finger = ((hw->z > 0) && priv->finger_state < FS_TOUCHED)?FS_TOUCHED
                :((hw->z ==0) && priv->finger_state > FS_UNTOUCHED)?FS_UNTOUCHED
                : priv->finger_state;
        
        return finger;
}

static int
GetTimeOut(ElantechPrivate *priv)
{
    ElantechSHM *para = priv->elapara;

    switch (priv->tap_state) {
    case TS_1:
    case TS_3:
    case TS_5:
        return para->tap_time;
    case TS_SINGLETAP:
        return para->click_time;
    case TS_2A:
        return para->single_tap_timeout;
    case TS_2B:
        return para->tap_time_2;
    case TS_4:
	/*para->locked_drag_time = para->drag_slider*1000.0;*/
        return (para->drag_slider*1000.0);
    default:
        return -1;                          /* No timeout */
    }
}

static int
clamp(int val, int min, int max)
{
    if (val < min)
	return min;
    else if (val < max)
	return val;
    else
	return max;
}

static void
SelectTapButton(ElantechPrivate *priv)
{
    TapEvent tap;

    if (priv->elapara->touchpad_off == 2) {
        priv->tap_button = 0;
        return;
    }

    switch (priv->tap_max_fingers) {
    case 1:
    default:
        DBG(7, ErrorF("no edge\n"));
        tap = F1_TAP;
        break;
    case 2:
        DBG(7, ErrorF("two finger tap\n"));
        tap = F2_TAP;
        break;
    case 3:
        DBG(7, ErrorF("three finger tap\n"));
        tap = F3_TAP;
        break;
    }

    priv->tap_button = priv->elapara->tap_action[tap];
    priv->tap_button = clamp(priv->tap_button, 0, ELA_MAX_BUTTONS);
}
static void
SetTapState(ElantechPrivate *priv, enum TapState tap_state, int millis)
{
    /*ElantechSHM *para = priv->elapara;*/
    DBG(7, ErrorF("SetTapState - %d -> %d (millis:%d)\n", priv->tap_state, tap_state, millis));
    switch (tap_state) {
    case TS_START:
        priv->tap_button_state = TBS_BUTTON_UP;
        priv->tap_max_fingers = 0;
        break;
    case TS_1:
        priv->tap_button_state = TBS_BUTTON_UP;
        break;
    case TS_2A:
        priv->tap_button_state = TBS_BUTTON_UP;
        break;
    case TS_2B:
        priv->tap_button_state = TBS_BUTTON_UP;
        break;
    case TS_3:
	
     	priv->tap_button_state = TBS_BUTTON_DOWN; 
        break;
    case TS_SINGLETAP:
        priv->tap_button_state = TBS_BUTTON_DOWN;
        priv->touch_on.millis = millis;
        break;
    default:
        break;
    }
    priv->tap_state = tap_state;
}

static void
SetMovingState(ElantechPrivate *priv, enum MovingState moving_state, int millis)
{
    DBG(7, ErrorF("SetMovingState - %d -> %d center at %d/%d (millis:%d)\n", priv->moving_state,
                  moving_state,priv->hwState.x, priv->hwState.y, millis));

    if (moving_state == MS_TRACKSTICK) {
        priv->trackstick_neutral_x = priv->hwState.x;
        priv->trackstick_neutral_y = priv->hwState.y;
    }
    priv->moving_state = moving_state;
}




static int
HandleTapProcessing(ElantechPrivate *priv, struct ElantechHwState *hw,
                    enum FingerState finger)
{
    ElantechSHM *para = priv->elapara;
    Bool touch,release,is_timeout,move;
    int timeleft,timeout;
    int delay = 1000000000;
    static int palmtrack_tap_flag=0;

    if(((7-para->palmtracking_slider) < para->w )&& para->drag_timeout_release &&
	 para->palmtracking_checkBox && palmtrack_tap_flag==0 && priv->has_pressure)
	  palmtrack_tap_flag=1;
    
    if((((7+4)-para->palmtracking_slider) < para->w ) && para->palmtracking_checkBox && priv->has_pressure)
	  return delay;
    
    touch = finger && !priv->finger_state;
    release = !finger && priv->finger_state;
    if(release)
	     palmtrack_tap_flag=0;
    move = ((priv->tap_max_fingers <=1) && 
	    ((abs(para->x - priv->touch_on.x) >= para->tap_move) ||
	     (abs(para->y - priv->touch_on.y) >= para->tap_move)||
             (para->tapping_checkBox == 0 || para->left || para->right) ||
	      palmtrack_tap_flag));
     
    if(touch) {
	  priv->touch_on.x = para->x;
	  priv->touch_on.y = para->y;
	  priv->touch_on.millis = hw->millis;
	  
    } else if(release) {
	  priv->touch_on.millis = hw->millis;
    }

        
    if(priv->tap_max_fingers < para->numFingers)
	    priv->tap_max_fingers = para->numFingers;
    timeout = GetTimeOut(priv);
    timeleft = TIME_DIFF(priv->touch_on.millis + timeout,hw->millis);
    is_timeout = timeleft <=0 ;
  
    para->locked_drags=!(para->lift_release_radioButton);
    
    if(para->drag_checkBox==1)
	para->single_tap_timeout=180;
    else
	para->single_tap_timeout=0;
   
#if 1	
restart:
    switch (priv->tap_state) {
	case TS_START:
		if (touch)
		      SetTapState(priv,TS_1,hw->millis);
		break;
	case TS_1:
		if (move) {
			SetMovingState(priv,MS_TOUCHPAD_RELATIVE,hw->millis);
			SetTapState(priv,TS_MOVE,hw->millis);
			 goto restart;
		} else if (is_timeout) {
			if (finger == FS_TOUCHED) {
				SetMovingState(priv,MS_TOUCHPAD_RELATIVE,hw->millis);
			} else if (finger == FS_PRESSED) {
				  SetMovingState(priv,MS_TRACKSTICK,hw->millis);
			}
			SetTapState(priv,TS_MOVE,hw->millis);
			goto restart;
		} else if (release) {
			SelectTapButton(priv);
			SetTapState(priv,TS_2A,hw->millis);
		}
		break;
	case TS_MOVE:
		if (move && priv->moving_state == MS_TRACKSTICK) {
			SetMovingState(priv,MS_TOUCHPAD_RELATIVE,hw->millis);
		}
		if (release) {
		    SetMovingState(priv,MS_FALSE,hw->millis);
		    SetTapState(priv,TS_START,hw->millis);
		}
		break;
	case TS_2A:
		if (touch){
			SetTapState(priv,TS_3,hw->millis);
		} else if(is_timeout){
			SetTapState(priv,TS_SINGLETAP,hw->millis);
		}
		break;
	case TS_2B:
		if (touch){
			SetTapState(priv,TS_3,hw->millis);
		} else if (is_timeout) {
			 SetTapState(priv, TS_START, hw->millis);
			  priv->tap_button_state = TBS_BUTTON_DOWN_UP;
		}
		break;
	case TS_SINGLETAP:
		if (touch) {
			SetTapState(priv,TS_1,hw->millis);
		} else if (is_timeout){
			SetTapState(priv,TS_START,hw->millis);
		}
		break;
	case TS_3:
		if (move) {
			SetMovingState(priv, MS_TOUCHPAD_RELATIVE, hw->millis);
			SetTapState(priv, TS_DRAG, hw->millis);
			goto restart;
		} else if (is_timeout) {
			if (finger == FS_TOUCHED) {
				 SetMovingState(priv, MS_TOUCHPAD_RELATIVE, hw->millis);
			} else if (finger == FS_PRESSED) {
				 SetMovingState(priv, MS_TRACKSTICK, hw->millis);
			}
			SetTapState(priv, TS_DRAG, hw->millis);
			goto restart;
		} else if (release) {
			SetTapState(priv, TS_2B, hw->millis);
		}
		break;
	case TS_DRAG:
		if (move){
			SetMovingState(priv, MS_TOUCHPAD_RELATIVE, hw->millis);
		}
		if (release) {
			SetMovingState(priv, MS_FALSE, hw->millis);
			if (para->locked_drags) {
				SetTapState(priv, TS_4, hw->millis);
			} else {
				SetTapState(priv, TS_START, hw->millis);
			}
		}
		break;
	case TS_4:
		if (is_timeout && para->tap_release_radioButton ) {
			 SetTapState(priv, TS_START, hw->millis);
			  goto restart;
		}
		if (touch){
			SetTapState(priv, TS_5, hw->millis);
		}
		
		break;
	case TS_5:
		if (is_timeout || move) {
			SetTapState(priv, TS_DRAG, hw->millis);
			goto restart;
		} else if (release) {			
			SetMovingState(priv, MS_FALSE, hw->millis);
			SetTapState(priv, TS_START, hw->millis);
		}		
		break;
}

    timeout = GetTimeOut(priv);
    if (timeout >= 0) {
        timeleft = TIME_DIFF(priv->touch_on.millis + timeout, hw->millis);
        delay = clamp(timeleft, 1, delay);
    }
#endif 
    return delay;


}

#define HIST(a) (priv->move_hist[((priv->hist_index - (a) + ELANTECH_MOVE_HISTORY) % ELANTECH_MOVE_HISTORY)])

static void
store_history(ElantechPrivate *priv, int x, int y, unsigned int millis)
{
    int idx = (priv->hist_index + 1) % ELANTECH_MOVE_HISTORY;
    priv->move_hist[idx].x = x;
    priv->move_hist[idx].y = y;
    priv->move_hist[idx].millis = millis;
    priv->hist_index = idx;
}

static double
estimate_delta(double x0, double x1, double x2, double x3)
{
    return x0 * 0.3 + x1 * 0.1 - x2 * 0.1 - x3 * 0.3;
}

static int
move_distance(int dx, int dy)
{
    return sqrt(SQR(dx) + SQR(dy));
}

static int
ComputeDeltas(ElantechPrivate *priv, struct ElantechHwState *hw,
	      int *dxP, int *dyP)
{
    ElantechSHM *para = priv->elapara;
    enum MovingState moving_state;
    int dist;
    double dx,dy;
    double speed,integral;
    int delay = 1000000000;
    
    dx = dy =0;
    
    moving_state = priv->moving_state;
    if (moving_state == MS_FALSE) {
	switch (priv->tap_state) {
	    case TS_MOVE:
	    case TS_DRAG:
		moving_state = MS_TOUCHPAD_RELATIVE;
		break;
	    case TS_1:
	    case TS_3:
	    case TS_5:
		if (para->numFingers == 1)
		      moving_state = MS_TOUCHPAD_RELATIVE;
		break;
	    default:
		break;
	}
      }
      
      if (moving_state && para->numFingers == 1 && para->detect_kbd_intr == 0) {  /*tom + */
	  delay = MIN(delay,13);
	  if (priv->count_packet_finger > 3 ) {
	      double tmpf;
	      double dtime = (hw->millis - HIST(0).millis)/1000.0;
	      
	      if (priv->moving_state == MS_TRACKSTICK) {
		dx = (para->x - priv->trackstick_neutral_x);
		dy = (para->y - priv->trackstick_neutral_y);

		dx = dx * dtime * para->trackstick_speed;
		dy = dy * dtime * para->trackstick_speed;
	      
	      } else if (moving_state == MS_TOUCHPAD_RELATIVE) {
		dx = estimate_delta(hw->x, HIST(0).x, HIST(1).x, HIST(2).x);
		dy = estimate_delta(hw->y, HIST(0).y, HIST(1).y, HIST(2).y);
		
	      }
	      
	       /* speed depending on distance/packet */
	    
	      dist = move_distance(dx, dy);
	      speed = dist * para->accl;
	      if (speed > para->max_speed) {  /* set max speed factor */
		speed = para->max_speed;
	      } else if (speed < para->min_speed) { /* set min speed factor */
		speed = para->min_speed;
	      }
	      
	       /* save the fraction, report the integer part */
	      tmpf = dx * speed + priv->frac_x;
	      priv->frac_x = modf(tmpf, &integral);
	      dx = integral;
	      tmpf = dy * speed  + priv->frac_y;
	      priv->frac_y = modf(tmpf, &integral);
	      dy = integral;
	     }   
	     priv->count_packet_finger++;
	   } else {
	      priv->count_packet_finger = 0;
	   }
	   *dxP = dx;
	   *dyP = dy;
	   store_history(priv, hw->x, hw->y, hw->millis);
          return delay;

}

/*
 * React on changes in the hardware state. This function is called every time
 * the hardware state changes. The return value is used to specify how many
 * milliseconds to wait before calling the function again if no state change
 * occurs.
 */

static int
HandleState(LocalDevicePtr local, struct ElantechHwState *hw)
{
    ElantechPrivate *priv = (ElantechPrivate *) (local->private);
    ElantechSHM *para = priv->elapara;
    int finger;
    int dx,dy,buttons,rep_buttons,id;
     struct ScrollData scroll;
   
#if 0
    static int tap_intr_one_time=0;
    static int tapping_checkBox_one_time=0;
    static int release_palmtrack_flag=0;
    static int palmtrack_tap_flag=0;
    static int tap_intr_save_tapdata=0;
    static int palmtrack_intr_save_tapdata=0;
#endif
   
    int change;
    
    int double_click,repeat_delay=0;
    int delay =  1000000000;
    int timeleft;
    
    
    /* update hardware state in share memory */
    para->x = hw->x;
    para->y = hw->y;
    para->z = hw->z;
    para->w = hw->w;
    
    para->x2_1 = hw->x2_1;
    para->y2_1 = hw->y2_1;
    para->x2_2 = hw->x2_2;
    para->y2_2 = hw->y2_2;
    para->x3_1 = hw->x3_1;
    para->y3_1 = hw->y3_1;
    
    para->numFingers = hw->numFingers;  
    para->MKY = hw->MKY;
    para->VF = hw->VF;
    
    para->left = hw->left;
    para->right = hw->right;
    para->middle = hw->middle;
    
    if (priv->smart_pad == 0){
	para->zoom_checkBox = 0;
	para->rotation_checkBox = 0;
	para->swipe_page_checkBox = 0;
	para->fast_key_checkBox = 0;	
    }

    if (priv->has_pressure == 0)
    {
	para->palmtracking_checkBox = 0;
    }

    if (priv->convergesture==0)
    {
	para->covergesture_checkBox = 0;
    }

   if (para->touchpad_off == 1)
	return delay;
	
   para->middle |= HandleMidButtonEmulation(priv, hw,&delay);
   
   double_click = FALSE;
   
   finger = ElantechDetectFinger(priv,hw);

    /* tap and drag detection */
    timeleft = HandleTapProcessing(priv, hw, finger);
    if (timeleft > 0)
	delay = MIN(delay, timeleft);

    HandleScrollProcessing(priv,hw->x, hw->y,&scroll);
    
    timeleft = ComputeDeltas(priv, hw, &dx, &dy);
    delay = MIN(delay, timeleft);
    
    rep_buttons = (para->leftright_button_repeat ? 0x60 : 0);

    if(para->new_struct == 0){
		buttons = ((para->left     ? 0x01 : 0) |
			    (para->middle   ? 0x02 : 0) |
			    (para->right    ? 0x04 : 0));
    } else {
		
		if(para->button_left_comboBox==0 && para->left)
			buttons=0x01;
		else if (para->button_right_comboBox==0 && para->right) 
			buttons=0x01;
		else if ( para->button_left_comboBox==1 && para->left)
			buttons=0x04;
		else if (para->button_right_comboBox==1 && para->right)
			buttons=0x04;
		else if ( para->button_left_comboBox==2 && para->left)
			buttons=0x02;
		else if (para->button_right_comboBox==2 && para->right)
			buttons=0x02;
		else 
			buttons = 0;

		if (para->button_checkBox==0)
			 buttons = 0;
    }

#if 1
    if (priv->tap_button > 0) {
	int tap_mask = 1 << (priv->tap_button - 1);
	if (priv->tap_button_state == TBS_BUTTON_DOWN_UP) {
	    if (tap_mask != (priv->lastButtons & tap_mask)) {
		xf86PostButtonEvent(local->dev, FALSE, priv->tap_button, TRUE, 0, 0);
		priv->lastButtons |= tap_mask;
	    }
	    priv->tap_button_state = TBS_BUTTON_UP;
	}
	if (priv->tap_button_state == TBS_BUTTON_DOWN)
	    buttons |= tap_mask;
    }
#endif
     if (dx || dy)
    	{
		if(para->can_edge_scroll_function==0)
        		xf86PostMotionEvent(local->dev, 0, 0, 2, dx, dy);
    		else
			para->can_edge_scroll_function=0;
   	}
	
    change = buttons ^ priv->lastButtons;
    while (change) {
	id = ffs(change); /* number of first set bit 1..32 is returned */
	change &= ~(1 << (id - 1));
	xf86PostButtonEvent(local->dev, FALSE, id, (buttons & (1 << (id - 1))), 0, 0);
    }
    
while (scroll.up-- > 0) 
    {
       	xf86PostButtonEvent(local->dev, FALSE, 4, !hw->up, 0, 0);
       	xf86PostButtonEvent(local->dev, FALSE, 4, hw->up, 0, 0);
    }
    while (scroll.down-- > 0) 
    {
        xf86PostButtonEvent(local->dev, FALSE, 5, !hw->down, 0, 0);
        xf86PostButtonEvent(local->dev, FALSE, 5, hw->down, 0, 0);
    }
    while (scroll.left-- > 0) 
    {
        xf86PostButtonEvent(local->dev, FALSE, 6, TRUE, 0, 0);
        xf86PostButtonEvent(local->dev, FALSE, 6, FALSE, 0, 0);
    }
    while (scroll.right-- > 0) 
    {
        xf86PostButtonEvent(local->dev, FALSE, 7, TRUE, 0, 0);
        xf86PostButtonEvent(local->dev, FALSE, 7, FALSE, 0, 0);
    }

    if (double_click) {
	int i;
	for (i = 0; i < 2; i++) {
	    xf86PostButtonEvent(local->dev, FALSE, 1, !hw->left, 0, 0);
	    xf86PostButtonEvent(local->dev, FALSE, 1, hw->left, 0, 0);
	}
    }
    repeat_delay = clamp (para->scroll_button_repeat,SBR_MIN,SBR_MAX);
    priv->repeatButtons = 0;
    priv->nextRepeat = 0;

    if (priv->repeatButtons) {
        timeleft = TIME_DIFF(priv->nextRepeat, hw->millis);
        if (timeleft > 0)
            delay = MIN(delay, timeleft);
        if (timeleft <= 0) {
            int change, id;
            change = priv->repeatButtons;
            while (change) {
                id = ffs(change);
                change &= ~(1 << (id - 1));
                xf86PostButtonEvent(local->dev, FALSE, id, FALSE, 0, 0);
                xf86PostButtonEvent(local->dev, FALSE, id, TRUE, 0, 0);
            }
            priv->nextRepeat = hw->millis + repeat_delay;
            delay = MIN(delay, repeat_delay);
        }
    }
    
     /* Save old values of some state variables */
    priv->finger_state = finger;
    priv->lastButtons = buttons;

    return delay;

}

static int
ControlProc(LocalDevicePtr local, xDeviceCtl * control)
{
    DBG(3, ErrorF("Control Proc called\n"));
    return Success;
}

static void
CloseProc(LocalDevicePtr local)
{
    DBG(3, ErrorF("Close Proc called\n"));
}

static int
SwitchMode(ClientPtr client, DeviceIntPtr dev, int mode)
{
    ErrorF("SwitchMode called\n");
    return Success;
}

static Bool
ConvertProc(LocalDevicePtr local,
	    int first,
	    int num,
	    int v0,
	    int v1,
	    int v2,
	    int v3,
	    int v4,
	    int v5,
	    int *x,
	    int *y)
{
    if (first != 0 || num != 2)
	return FALSE;

    *x = v0;
    *y = v1;

    return TRUE;
}

static void
ReadDevDimensions(LocalDevicePtr local)
{
    ElantechPrivate *priv = (ElantechPrivate *)local->private;
    
    if (priv->proto_ops->ReadDevDimensions)
	priv->proto_ops->ReadDevDimensions(local);
}

static Bool
QueryHardware(LocalDevicePtr local)
{
    ElantechPrivate *priv = (ElantechPrivate *) local->private;
    /*ElantechSHM *para = priv->elapara;*/
    
    priv->comm.protoBufTail = 0;
    
    if (priv->proto_ops->QueryHardware(local)) {
    } else {
	xf86Msg(X_PROBED, "%s: no supported touchpad found\n",local->name);
	priv->proto_ops->DeviceOffHook(local);
    }
    
    return TRUE;
 }
